"""
.. Utilities
"""

from __future__ import annotations

__all__ = (
    "get_cell_size",
    "get_terminal_name_version",
    "get_terminal_size",
    "lock_tty",
    "read_tty_all",
    "write_tty",
)

import os
import sys
import warnings
from array import array
from functools import wraps
from multiprocessing import Array, Process, Queue as mp_Queue, RLock as mp_RLock
from operator import floordiv
from queue import Empty, Queue
from shutil import get_terminal_size as _get_terminal_size
from threading import RLock
from time import monotonic
from types import FunctionType
from typing import Any, Callable, Optional, Tuple, Union

from . import ctlseqs
from .exceptions import TermImageWarning

# import logging

OS_IS_UNIX: bool
try:
    import fcntl
    import termios
    from select import select
except ImportError:
    OS_IS_UNIX = False
else:
    OS_IS_UNIX = True

# NOTE: Any cached feature using a query should have it's cache invalidated in
# `term_image.enable_queries()`.

# Decorator Classes


class ClassInstanceMethod(classmethod):
    """A method which when invoked via the owner, behaves like a class method
    and when invoked via an instance, behaves like an instance method.
    """

    def __init__(
        self, f_owner: FunctionType, f_instance: Optional[FunctionType] = None
    ):
        super().__init__(f_owner)
        self.f_owner = f_owner
        self.f_instance = f_instance

    def __get__(self, instance, owner=None):
        if instance:
            return self.f_instance.__get__(instance, owner)
        else:
            return super().__get__(instance, owner)

    def classmethod(self, function: FunctionType) -> ClassInstanceMethod:
        return type(self)(function, self.f_instance)

    def instancemethod(self, function: FunctionType) -> ClassInstanceMethod:
        return type(self)(self.f_owner, function)


class ClassPropertyBase(property):
    """Base class for owner properties that also have a counterpart/shadow on the
    instance.
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        super().__init__(fget, fset, fdel, doc)
        # `property` doesn't set `__doc__`, probably cos the subclass' `__doc__`
        # attribute overrides its `__doc__` descriptor.
        super().__setattr__("__doc__", doc or fget.__doc__)


class ClassInstanceProperty(ClassPropertyBase):
    """A an instance-specific counterpart of a property of the owner.

    Operation on the owner is actually implemented by a property defined on the
    owner's metaclass. This class is only for the sake of ease of documentation
    without having to bother the user about metaclasses.
    """


class ClassProperty(ClassPropertyBase):
    """A read-only shadow of a property of the owner.

    Operation on the owner is actually implemented by a property defined on the
    owner's metaclass. This class is only for the sake of ease of documentation
    without having to bother the user about metaclasses.
    """


# Decorator Functions


def no_redecorate(decor: FunctionType) -> FunctionType:
    """Prevents a decorator from re-decorating objects.

    Args:
        decor: The decorator to be wrapped.
    """
    if getattr(decor, "_no_redecorate_", False):
        return decor

    @wraps(decor)
    def no_redecorate_wrapper(obj, *args, **kwargs):
        if not getattr(obj, f"_{decor.__name__}_", False):
            obj = decor(obj, *args, **kwargs)
            setattr(obj, f"_{decor.__name__}_", True)
        return obj

    no_redecorate_wrapper._no_redecorate_ = True
    return no_redecorate_wrapper


@no_redecorate
def cached(func: FunctionType) -> FunctionType:
    """Enables return value caching.

    Args:
        func: The function to be wrapped.

    An *_invalidate_cache* function is set as an attribute of the returned wrapper
    which when called clears the cache, so that the next call actually calls the
    wrapped function, no matter the value of *_cached*.

    NOTE:
        It's thread-safe, i.e there is no race condition between calls to the same
        decorated object across threads of the same process.

        Only works when function arguments, if any, are hashable.
    """

    @wraps(func)
    def cached_wrapper(*args, **kwargs):
        arguments = (args, tuple(kwargs.items()))
        with lock:
            try:
                return cache[arguments]
            except KeyError:
                return cache.setdefault(arguments, func(*args, **kwargs))

    def invalidate():
        with lock:
            cache.clear()

    cache = {}
    lock = RLock()
    cached_wrapper._invalidate_cache = invalidate

    return cached_wrapper


@no_redecorate
def lock_tty(func: FunctionType) -> FunctionType:
    """Synchronizes access to the :term:`active terminal`.

    Args:
        func: The function to be wrapped.

    When a decorated function is called, a re-entrant lock is acquired by the current
    process or thread and released after the call, such that any other decorated
    function called within another thread or subprocess waits until the lock is
    fully released (i.e has been released as many times as acquired) by the current
    process or thread.

    NOTE:
        It works across parent-/sub-processes, started directly or indirectly via
        :py:class:`multiprocessing.Process` (or a subclass of it), and their threads,
        provided :py:mod:`multiprocessing.synchronize` is supported on the host
        platform. Otherwise, a warning is issued when starting a subprocess.

    WARNING:
        If :py:mod:`multiprocessing.synchronize` is supported and a subprocess is
        started within a call (possibly recursive) to a decorated function, the thread
        in which that occurs will be out of sync until that call returns.
        Hence, avoid starting a subprocess within a decorated function.
    """

    @wraps(func)
    def lock_tty_wrapper(*args, **kwargs):
        # If a thread reaches this point while the lock is being changed
        # (the old lock has been acquired but hasn't been changed), after the lock has
        # been changed and the former lock is released, the waiting thread will acquire
        # the old lock making it to be out of sync.
        # Hence the second expression, which allows such a thread to acquire the new
        # lock and be in sync.
        # NB: Multiple expressions are processed as multiple nested with statements.
        with _tty_lock, _tty_lock:
            # logging.debug(f"{func.__name__} acquired TTY lock", stacklevel=3)
            return func(*args, **kwargs)

    if func.__module__.startswith("term_image") and func.__doc__ is not None:
        sync_doc = """

        IMPORTANT:
            Synchronized with :py:func:`~term_image.utils.lock_tty`.
        """

        last_line = func.__doc__.rpartition("\n")[2]
        indent = " " * (len(last_line) - len(last_line.lstrip()))
        lock_tty_wrapper.__doc__ = func.__doc__.rstrip() + "\n".join(
            line.replace(" " * 8, indent, 1) for line in sync_doc.splitlines()
        )

    return lock_tty_wrapper


@no_redecorate
def terminal_size_cached(func: FunctionType) -> FunctionType:
    """Enables return value caching based on the size of the :term:`active terminal`.

    Args:
        func: The function to be wrapped.

    If the terminal size is the same as for the last call, the last return value is
    returned. Otherwise the wrapped function is called and the new return value is
    stored and returned.

    An *_invalidate_terminal_size_cache* function is also set as an attribute of the
    returned wrapper which when called clears the cache, so that the next call actually
    calls the wrapped function.

    NOTE:
        It's thread-safe, i.e there is no race condition between calls to the same
        decorated callable across threads of the same process.
    """

    @wraps(func)
    def terminal_size_cached_wrapper(*args, **kwargs):
        with lock:
            ts = get_terminal_size()
            if not cache or ts != cache[1]:
                cache[:] = [func(*args, **kwargs), ts]
        return cache[0]

    def invalidate():
        with lock:
            cache.clear()

    cache = []
    lock = RLock()
    terminal_size_cached_wrapper._invalidate_terminal_size_cache = invalidate

    return terminal_size_cached_wrapper


@no_redecorate
def unix_tty_only(func: FunctionType) -> FunctionType:
    """Disable invocation of a function on a non-unix-like platform or when there is no
    :term:`active terminal`.

    Args:
        func: The function to be wrapped.
    """

    @wraps(func)
    def unix_only_wrapper(*args, **kwargs):
        return _tty_fd and func(*args, **kwargs)

    unix_only_wrapper.__doc__ += """
    NOTE:
        Currently works on UNIX only, returns ``None`` on any other platform or when
        there is no :term:`active terminal`.
    """

    return unix_only_wrapper


# Non-decorators


def arg_type_error(arg: str, value: Any) -> TypeError:
    return TypeError(f"Invalid type for {arg!r} (got: {type(value).__name__})")


def arg_value_error(arg: str, value: Any) -> ValueError:
    return ValueError(f"Invalid value for {arg!r} (got: {value!r})")


def arg_value_error_msg(msg: str, value: Any) -> ValueError:
    return ValueError(f"{msg} (got: {value!r})")


def arg_value_error_range(arg: str, value: Any, got_extra: str = "") -> ValueError:
    return ValueError(
        f"{arg!r} out of range (got: {value!r}, {got_extra})"
        if got_extra
        else f"{arg!r} out of range (got: {value!r})"
    )


def clear_queue(queue: Union[Queue, mp_Queue]):
    """Purges the given queue"""
    while True:
        try:
            queue.get(timeout=0.005)
        except Empty:
            break


def color(
    text: str, fg: Tuple[int] = (), bg: Tuple[int] = (), *, end: bool = False
) -> str:
    """Prepends *text* with 24-bit color escape codes for the given foreground and/or
    background RGB values, optionally ending with the color reset sequence.

    Args:
        text: String to be color-coded.
        fg: Foreground color.
        bg: Background color.
        end: If ``True``, the color reset sequence is appended to the returned string.

    Returns:
        The color-coded string.

    The color code is omitted for any of *fg* or *bg* that is empty.
    """
    return (ctlseqs.SGR_FG_RGB * bool(fg) + ctlseqs.SGR_BG_RGB * bool(bg) + "%s") % (
        *fg,
        *bg,
        text,
    ) + ctlseqs.SGR_NORMAL * end


@unix_tty_only
def get_cell_size() -> Optional[Tuple[int, int]]:
    """Returns the current size of a character cell in the :term:`active terminal`.

    Returns:
        The cell size in pixels or `None` if undetermined.

    The speed of this implementation is almost entirely dependent on the terminal; the
    method it supports and its response time if it has to be queried.
    """
    # If a thread reaches this point while the lock is being changed
    # (the old lock has been acquired but hasn't been changed), after the lock has
    # been changed and the former lock is released, the waiting thread will acquire
    # the old lock making it to be out of sync.
    # Hence the second expression, which allows such a thread to acquire the new
    # lock and be in sync.
    # NB: Multiple expressions are processed as multiple nested with statements.
    with _cell_size_lock, _cell_size_lock:
        ts = get_terminal_size()
        if ts == tuple(_cell_size_cache[:2]):
            size = tuple(_cell_size_cache[2:])
            return None if 0 in size else size

        size = text_area_size = None

        # First try ioctl
        buf = array("H", [0, 0, 0, 0])
        try:
            if not fcntl.ioctl(_tty_fd, termios.TIOCGWINSZ, buf):
                text_area_size = tuple(buf[2:])
                if 0 in text_area_size:
                    text_area_size = None
        except OSError:
            pass

        if not text_area_size:
            # Then XTWINOPS
            # The last sequence is to speed up the entire query since most (if not all)
            # terminals should support it and most terminals treat queries as FIFO
            response = query_terminal(
                ctlseqs.CELL_SIZE_PX_b + ctlseqs.TEXT_AREA_SIZE_PX_b + ctlseqs.DA1_b,
                more=lambda s: not s.endswith(b"c"),
            )
            if response:
                cell_size_match = ctlseqs.CELL_SIZE_PX_re.match(response.decode())
                if not cell_size_match:
                    text_area_size_match = ctlseqs.TEXT_AREA_SIZE_PX_re.match(
                        response.decode()
                    )

                # XTWINOPS specifies (height, width)
                if cell_size_match:
                    size = tuple(map(int, cell_size_match.groups()))[::-1]
                elif text_area_size_match:
                    text_area_size = tuple(  # fmt: skip
                        map(int, text_area_size_match.groups())
                    )[::-1]

                    # Termux seems to respond with (height / 2, width), though the
                    # values are incorrect as they change with different zoom levels
                    # but still always give a reasonable (almost always the same)
                    # cell size and ratio.
                    if os.environ.get("SHELL", "").startswith("/data/data/com.termux/"):
                        text_area_size = (text_area_size[0], text_area_size[1] * 2)

        if text_area_size and _swap_win_size:
            text_area_size = text_area_size[::-1]
        size = size or (
            tuple(map(floordiv, text_area_size, ts)) if text_area_size else (0, 0)
        )
        _cell_size_cache[:] = ts + size

        return None if 0 in size else size


@cached
def get_fg_bg_colors(
    *, hex: bool = False
) -> Tuple[
    Union[None, str, Tuple[int, int, int]], Union[None, str, Tuple[int, int, int]]
]:
    """Returns the default FG and BG colors of the :term:`active terminal`.

    Returns:
        For each color:

        * an RGB 3-tuple, if *hex* is ``False``
        * an RGB hex string if *hex* is ``True``
        * ``None`` if undetermined
    """
    # The terminal's response to the queries is not read all at once
    with _tty_lock, _tty_lock:  # See the comment in `lock_tty_wrapper()`
        response = query_terminal(
            # Not all terminals (e.g VTE-based) support multiple queries in one escape
            # sequence, hence the separate sequences for FG and BG
            ctlseqs.TEXT_FG_QUERY_b + ctlseqs.TEXT_BG_QUERY_b + ctlseqs.DA1_b,
            # The response might contain a "c"; can't stop reading at "c"
            lambda s: not s.endswith(ctlseqs.CSI_b),
        )
        if _queries_enabled:
            read_tty()  # The rest of the response to DA1

    fg = bg = None
    if response:
        for c, spec in ctlseqs.RGB_SPEC_re.findall(response.decode()):
            if c == "10":
                fg = ctlseqs.x_parse_color(spec)
            elif c == "11":
                bg = ctlseqs.x_parse_color(spec)

    return tuple(
        rgb and ("#" + ("{:02x}" * 3).format(*rgb) if hex else rgb) for rgb in (fg, bg)
    )


@cached
def get_terminal_name_version() -> Tuple[Optional[str], Optional[str]]:
    """Returns the name and version of the :term:`active terminal`, if available.

    Returns:
        A 2-tuple, ``(name, version)``. If either is not available, returns ``None``
        in its place.
    """
    # The terminal's response to the queries is not read all at once
    with _tty_lock, _tty_lock:  # See the comment in `lock_tty_wrapper()`
        # Terminal name/version query + terminal attribute query
        # The latter is to speed up the entire query since most (if not all)
        # terminals should support it and most terminals treat queries as FIFO
        response = query_terminal(
            ctlseqs.XTVERSION_b + ctlseqs.DA1_b,
            # The response might contain a "c"; can't stop reading at "c"
            lambda s: not s.endswith(ctlseqs.CSI_b),
        )
        if _queries_enabled:
            read_tty()  # The rest of the response to DA1

    match = response and ctlseqs.XTVERSION_re.match(response.decode())
    name, version = (
        match.groups()
        if match
        else map(os.environ.get, ("TERM_PROGRAM", "TERM_PROGRAM_VERSION"))
    )

    return (name and name.lower(), version)


def get_terminal_size() -> os.terminal_size:
    """Returns the current size of the :term:`active terminal`.

    Returns:
        The terminal size in columns and lines.

    NOTE:
        This implementation is quite different from :py:func:`shutil.get_terminal_size`
        and :py:func:`os.get_terminal_size` in that it:

        - gives the correct size of the :term:`active terminal` even when output is
          redirected, in most cases
        - gives different results in certain situations
        - is what this library works with
    """
    if _tty_fd:
        # faster and gives correct results when output is redirected
        try:
            size = os.get_terminal_size(_tty_fd)
        except OSError:
            size = None
    else:
        size = None

    return size or _get_terminal_size()


@unix_tty_only
@lock_tty
def query_terminal(
    request: bytes, more: Callable[[bytearray], bool], timeout: float = None
) -> Optional[bytes]:
    """Sends a query to the :term:`active terminal` and returns the response.

    Args:
        more: A callable, which when passed the response received so far, returns a
          boolean indicating if the response is incomplete or not. If it returns:

          * ``True``, more response is waited for.
          * ``False``, the received response is returned immediately.

        timeout: Time limit for awaiting a response from the terminal, in seconds
          (infinite if negative).

          If not given or ``None``, the value set by
          :py:func:`~term_image.set_query_timeout`
          (or :py:data:`~term_image.DEFAULT_QUERY_TIMEOUT` if never set) is used.

    Returns:
        `None` if queries are disabled (via :py:func:`~term_image.disable_queries`),
        else the terminal's response (empty, if no response is received after
        *timeout* is up).

    ATTENTION:
        Any unread input is discarded before the query. If the input might be needed,
        it can be read using :py:func:`read_tty()` before calling this function.
    """
    if not _queries_enabled:
        return None

    old_attr = termios.tcgetattr(_tty_fd)
    new_attr = termios.tcgetattr(_tty_fd)
    new_attr[3] &= ~termios.ECHO  # Disable input echo
    try:
        termios.tcsetattr(_tty_fd, termios.TCSAFLUSH, new_attr)
        write_tty(request)
        return read_tty(more, timeout or _query_timeout)
    finally:
        termios.tcsetattr(_tty_fd, termios.TCSANOW, old_attr)


@unix_tty_only
@lock_tty
def read_tty(
    more: Callable[[bytearray], bool] = lambda _: True,
    timeout: Optional[float] = None,
    min: int = 0,
    *,
    echo: bool = False,
) -> bytes:
    """Reads input directly from the :term:`active terminal` with/without blocking.

    Args:
        more: A callable, which when passed the input received so far, as a `bytearray`
          object, returns a boolean. If it returns:

          * ``True``, more input is waited for.
          * ``False``, the input received so far is returned immediately.

        timeout: Time limit for awaiting input, in seconds.
        min: Causes to block until at least the given number of bytes have been read.
        echo: If ``True``, any input while waiting is printed unto the screen.
          Any input before or after calling this function is not affected.

    Returns:
        The input read (empty, if *min* == ``0`` (default) and no input is received
        before *timeout* is up).

    If *timeout* is ``None`` (default), all available input is read without blocking.

    If *timeout* is not ``None`` and:

      * *timeout* < ``0``, it's infinite.
      * *min* > ``0``, input is waited for until at least *min* bytes have been read.

        After *min* bytes have been read, the following points apply with *timeout*
        being the leftover of the original *timeout*, if not yet used up.

      * *more* is not given, input is read or waited for until *timeout* is up.
      * *more* is given, input is read or waited for until ``more(input)`` returns
        ``False`` or *timeout* is up.

    Upon return or interruption, the :term:`active terminal` is **immediately** restored
    to the state in which it was met.
    """
    old_attr = termios.tcgetattr(_tty_fd)
    new_attr = termios.tcgetattr(_tty_fd)
    new_attr[3] &= ~termios.ICANON  # Disable canonical mode
    new_attr[6][termios.VTIME] = 0  # Never block based on time
    if echo:
        new_attr[3] |= termios.ECHO  # Enable input echo
    else:
        new_attr[3] &= ~termios.ECHO  # Disable input echo
    # Block until *min* bytes are read, when *timeout* is not `None`.
    new_attr[6][termios.VMIN] = 0 if timeout is None else min

    input = bytearray()
    try:
        termios.tcsetattr(_tty_fd, termios.TCSANOW, new_attr)
        r, w, x = [_tty_fd], [], []

        if timeout is None:
            # VMIN=0 does not work as expected on some platforms when there's no input
            while select(r, w, x, 0.0)[0]:
                input.extend(os.read(_tty_fd, 100))
        else:
            start = monotonic()
            if min > 0:
                input.extend(os.read(_tty_fd, min))

                # Don't block based on based on amount of bytes anymore
                new_attr[6][termios.VMIN] = 0
                termios.tcsetattr(_tty_fd, termios.TCSANOW, new_attr)

            duration = monotonic() - start
            while (timeout < 0 or duration < timeout) and more(input):
                # Reduces CPU usage
                # Also, VMIN=0 does not work on some platforms when there's no input
                if select(r, w, x, None if timeout < 0 else timeout - duration)[0]:
                    input.extend(os.read(_tty_fd, 1))
                duration = monotonic() - start
            # logging.debug(duration)
    finally:
        termios.tcsetattr(_tty_fd, termios.TCSANOW, old_attr)

    return bytes(input)


@unix_tty_only
def read_tty_all() -> bytes:
    """Reads all available input directly from the :term:`active terminal` **without
    blocking**.

    Returns:
        The input read.

    IMPORTANT:
        Synchronized with :py:func:`~term_image.utils.lock_tty`.
    """
    return read_tty()


@unix_tty_only
@lock_tty
def write_tty(data: bytes) -> None:
    """Writes to the :term:`active terminal` and waits until complete transmission.

    Args:
        data: Data to be written.
    """
    os.write(_tty_fd, data)
    try:
        termios.tcdrain(_tty_fd)
    except termios.error:  # "Permission denied" on some platforms e.g Termux
        pass


def _process_start_wrapper(self, *args, **kwargs):
    global _tty_lock, _cell_size_cache, _cell_size_lock

    # Ensure a lock is not acquired by another process/thread before changing it.
    # The only case in which this is useless is when the owner thread is the
    # one starting a process. In such a situation, the owner thread will be partially
    # (may acquire the new lock in a nested call while still holding the old lock)
    # out of sync until it has fully released the old lock.

    with _tty_lock:
        if isinstance(_tty_lock, _rlock_type):
            try:
                self._tty_lock = _tty_lock = mp_RLock()
            except ImportError:
                self._tty_lock = None
                warnings.warn(
                    "Multi-process synchronization is not supported on this platform!\n"
                    "Hence, if any subprocess will be writing/reading to/from the "
                    "active terminal, it may be unsafe to use any features requiring"
                    "terminal queries.\n"
                    "See https://term-image.readthedocs.io/en/stable/guide/concepts"
                    ".html#terminal-queries\n"
                    "If any related issues occur, it's advisable to disable queries "
                    "using `term_image.disable_queries()`.\n"
                    "Simply set an 'ignore' filter for this warning (before starting "
                    "any subprocess) if not using any of the affected features.",
                    TermImageWarning,
                )
        else:
            self._tty_lock = _tty_lock

    with _cell_size_lock:
        if isinstance(_cell_size_lock, _rlock_type):
            try:
                self._cell_size_cache = _cell_size_cache = Array("i", _cell_size_cache)
                _cell_size_lock = _cell_size_cache.get_lock()
            except ImportError:
                self._cell_size_cache = None
        else:
            self._cell_size_cache = _cell_size_cache

    return _process_start_wrapper.__wrapped__(self, *args, **kwargs)


def _process_run_wrapper(self, *args, **kwargs):
    global _tty_lock, _cell_size_cache, _cell_size_lock

    if self._tty_lock:
        _tty_lock = self._tty_lock
    if self._cell_size_cache:
        _cell_size_cache = self._cell_size_cache
        _cell_size_lock = _cell_size_cache.get_lock()

    return _process_run_wrapper.__wrapped__(self, *args, **kwargs)


# Private internal variables
_query_timeout = 0.1
_queries_enabled: bool = True
_swap_win_size: bool = False
_tty_fd: Optional[int] = None
_tty_lock = RLock()
_cell_size_cache = [0] * 4
_cell_size_lock = RLock()
_rlock_type = type(_tty_lock)

if OS_IS_UNIX:
    for stream in ("out", "in", "err"):  # In order of priority
        try:
            _tty_fd = os.ttyname(getattr(sys, f"__std{stream}__").fileno())
            break
        except (OSError, AttributeError):
            pass
    else:
        try:
            _tty_fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
        except OSError:
            warnings.warn(
                "It seems this process is not running within a terminal. "
                "Hence, some features will behave differently or be disabled.\n"
                "See https://term-image.readthedocs.io/en/stable/guide/concepts"
                ".html#active-terminal\n"
                "Any filter for this warning must be set before loading `term_image`, "
                "using `UserWarning` with the warning message (since "
                "`TermImageWarning` won't be available).",
                TermImageWarning,
            )
    if _tty_fd:
        if isinstance(_tty_fd, str):
            _tty_fd = os.open(_tty_fd, os.O_RDWR)

        Process.start = wraps(Process.start)(_process_start_wrapper)
        Process.run = wraps(Process.run)(_process_run_wrapper)

        # Shouldn't be needed since we're getting our own separate file descriptors
        # but the validity of the assumed safety is still under probation
        """
        for name, value in vars(termios).items():
            if isinstance(value, BuiltinFunctionType):
                setattr(termios, name, lock_tty(value))
        """

XTERM_256_PALETTE = (
    0, 0, 0,
    205, 0, 0,
    0, 205, 0,
    205, 205, 0,
    0, 0, 238,
    205, 0, 205,
    0, 205, 205,
    229, 229, 229,
    127, 127, 127,
    255, 0, 0,
    0, 255, 0,
    255, 255, 0,
    92, 92, 255,
    255, 0, 255,
    0, 255, 255,
    255, 255, 255,
    0, 0, 0,
    0, 0, 95,
    0, 0, 135,
    0, 0, 175,
    0, 0, 215,
    0, 0, 255,
    0, 95, 0,
    0, 95, 95,
    0, 95, 135,
    0, 95, 175,
    0, 95, 215,
    0, 95, 255,
    0, 135, 0,
    0, 135, 95,
    0, 135, 135,
    0, 135, 175,
    0, 135, 215,
    0, 135, 255,
    0, 175, 0,
    0, 175, 95,
    0, 175, 135,
    0, 175, 175,
    0, 175, 215,
    0, 175, 255,
    0, 215, 0,
    0, 215, 95,
    0, 215, 135,
    0, 215, 175,
    0, 215, 215,
    0, 215, 255,
    0, 255, 0,
    0, 255, 95,
    0, 255, 135,
    0, 255, 175,
    0, 255, 215,
    0, 255, 255,
    95, 0, 0,
    95, 0, 95,
    95, 0, 135,
    95, 0, 175,
    95, 0, 215,
    95, 0, 255,
    95, 95, 0,
    95, 95, 95,
    95, 95, 135,
    95, 95, 175,
    95, 95, 215,
    95, 95, 255,
    95, 135, 0,
    95, 135, 95,
    95, 135, 135,
    95, 135, 175,
    95, 135, 215,
    95, 135, 255,
    95, 175, 0,
    95, 175, 95,
    95, 175, 135,
    95, 175, 175,
    95, 175, 215,
    95, 175, 255,
    95, 215, 0,
    95, 215, 95,
    95, 215, 135,
    95, 215, 175,
    95, 215, 215,
    95, 215, 255,
    95, 255, 0,
    95, 255, 95,
    95, 255, 135,
    95, 255, 175,
    95, 255, 215,
    95, 255, 255,
    135, 0, 0,
    135, 0, 95,
    135, 0, 135,
    135, 0, 175,
    135, 0, 215,
    135, 0, 255,
    135, 95, 0,
    135, 95, 95,
    135, 95, 135,
    135, 95, 175,
    135, 95, 215,
    135, 95, 255,
    135, 135, 0,
    135, 135, 95,
    135, 135, 135,
    135, 135, 175,
    135, 135, 215,
    135, 135, 255,
    135, 175, 0,
    135, 175, 95,
    135, 175, 135,
    135, 175, 175,
    135, 175, 215,
    135, 175, 255,
    135, 215, 0,
    135, 215, 95,
    135, 215, 135,
    135, 215, 175,
    135, 215, 215,
    135, 215, 255,
    135, 255, 0,
    135, 255, 95,
    135, 255, 135,
    135, 255, 175,
    135, 255, 215,
    135, 255, 255,
    175, 0, 0,
    175, 0, 95,
    175, 0, 135,
    175, 0, 175,
    175, 0, 215,
    175, 0, 255,
    175, 95, 0,
    175, 95, 95,
    175, 95, 135,
    175, 95, 175,
    175, 95, 215,
    175, 95, 255,
    175, 135, 0,
    175, 135, 95,
    175, 135, 135,
    175, 135, 175,
    175, 135, 215,
    175, 135, 255,
    175, 175, 0,
    175, 175, 95,
    175, 175, 135,
    175, 175, 175,
    175, 175, 215,
    175, 175, 255,
    175, 215, 0,
    175, 215, 95,
    175, 215, 135,
    175, 215, 175,
    175, 215, 215,
    175, 215, 255,
    175, 255, 0,
    175, 255, 95,
    175, 255, 135,
    175, 255, 175,
    175, 255, 215,
    175, 255, 255,
    215, 0, 0,
    215, 0, 95,
    215, 0, 135,
    215, 0, 175,
    215, 0, 215,
    215, 0, 255,
    215, 95, 0,
    215, 95, 95,
    215, 95, 135,
    215, 95, 175,
    215, 95, 215,
    215, 95, 255,
    215, 135, 0,
    215, 135, 95,
    215, 135, 135,
    215, 135, 175,
    215, 135, 215,
    215, 135, 255,
    215, 175, 0,
    215, 175, 95,
    215, 175, 135,
    215, 175, 175,
    215, 175, 215,
    215, 175, 255,
    215, 215, 0,
    215, 215, 95,
    215, 215, 135,
    215, 215, 175,
    215, 215, 215,
    215, 215, 255,
    215, 255, 0,
    215, 255, 95,
    215, 255, 135,
    215, 255, 175,
    215, 255, 215,
    215, 255, 255,
    255, 0, 0,
    255, 0, 95,
    255, 0, 135,
    255, 0, 175,
    255, 0, 215,
    255, 0, 255,
    255, 95, 0,
    255, 95, 95,
    255, 95, 135,
    255, 95, 175,
    255, 95, 215,
    255, 95, 255,
    255, 135, 0,
    255, 135, 95,
    255, 135, 135,
    255, 135, 175,
    255, 135, 215,
    255, 135, 255,
    255, 175, 0,
    255, 175, 95,
    255, 175, 135,
    255, 175, 175,
    255, 175, 215,
    255, 175, 255,
    255, 215, 0,
    255, 215, 95,
    255, 215, 135,
    255, 215, 175,
    255, 215, 215,
    255, 215, 255,
    255, 255, 0,
    255, 255, 95,
    255, 255, 135,
    255, 255, 175,
    255, 255, 215,
    255, 255, 255,
    8, 8, 8,
    18, 18, 18,
    28, 28, 28,
    38, 38, 38,
    48, 48, 48,
    58, 58, 58,
    68, 68, 68,
    78, 78, 78,
    88, 88, 88,
    98, 98, 98,
    108, 108, 108,
    118, 118, 118,
    128, 128, 128,
    138, 138, 138,
    148, 148, 148,
    158, 158, 158,
    168, 168, 168,
    178, 178, 178,
    188, 188, 188,
    198, 198, 198,
    208, 208, 208,
    218, 218, 218,
    228, 228, 228,
    238, 238, 238,
)  # fmt: skip



""".. Image widgets for TUI frameworks"""

from __future__ import annotations

__all__ = []

try:
    import urwid
except ImportError:
    pass
else:
    from .urwid import UrwidImage, UrwidImageCanvas, UrwidImageScreen

    del urwid
    __all__ += ["UrwidImage", "UrwidImageCanvas", "UrwidImageScreen"]



"""
.. Core Library Definitions
"""

from __future__ import annotations

__all__ = (
    "auto_image_class",
    "AutoImage",
    "from_file",
    "from_url",
    "ImageSource",
    "Size",
    "BaseImage",
    "TextImage",
    "BlockImage",
    "GraphicsImage",
    "ITerm2Image",
    "KittyImage",
    "ImageIterator",
)

import os
from typing import Optional, Union

import PIL

from .block import BlockImage
from .common import (
    BaseImage,
    GraphicsImage,
    ImageIterator,
    ImageMeta,
    ImageSource,
    Size,
    TextImage,
)
from .iterm2 import ITerm2Image
from .kitty import KittyImage


def auto_image_class() -> ImageMeta:
    """Selects the image render style that best suits the current terminal emulator.

    Returns:
        An image class (a subclass of :py:class:`BaseImage`).
    """
    for cls in _styles:
        if cls.is_supported():
            break
    return cls


def AutoImage(
    image: PIL.Image.Image,
    *,
    width: Optional[int] = None,
    height: Optional[int] = None,
) -> BaseImage:
    """Creates an image instance from a PIL image instance.

    Returns:
        An instance of the automatically selected image render style (as returned by
        :py:func:`auto_image_class`).

    Same arguments and raised exceptions as the :py:class:`BaseImage` class constructor.
    """
    return auto_image_class()(image, width=width, height=height)


def from_file(
    filepath: Union[str, os.PathLike],
    **kwargs: Union[None, int],
) -> BaseImage:
    """Creates an image instance from an image file.

    Returns:
        An instance of the automatically selected image render style (as returned by
        :py:func:`auto_image_class`).

    Same arguments and raised exceptions as :py:meth:`BaseImage.from_file`.
    """
    return auto_image_class().from_file(filepath, **kwargs)


def from_url(
    url: str,
    **kwargs: Union[None, int],
) -> BaseImage:
    """Creates an image instance from an image URL.

    Returns:
        An instance of the automatically selected image render style (as returned by
        :py:func:`auto_image_class`).

    Same arguments and raised exceptions as :py:meth:`BaseImage.from_url`.
    """
    return auto_image_class().from_url(url, **kwargs)


# In order of preference, based on image quality and style performance/functionality
_styles = (KittyImage, ITerm2Image, BlockImage)



from __future__ import annotations

__all__ = ("KittyImage",)

import io
import re
import sys
from base64 import standard_b64encode
from dataclasses import asdict, dataclass
from typing import Any, Dict, Generator, Optional, Set, Tuple, Union
from zlib import compress, decompress

import PIL

from .. import ctlseqs

# These sequences are used during performance-critical operations that occur often
from ..ctlseqs import (
    CURSOR_FORWARD,
    ERASE_CHARS,
    KITTY_DELETE_CURSOR,
    KITTY_TRANSMISSION,
)
from ..utils import (
    arg_type_error,
    arg_value_error_msg,
    arg_value_error_range,
    get_terminal_name_version,
    query_terminal,
    write_tty,
)
from .common import GraphicsImage

# Constants for render methods
LINES = "lines"
WHOLE = "whole"


class KittyImage(GraphicsImage):
    """A render style using the Kitty terminal graphics protocol.

    See :py:class:`GraphicsImage` for the complete description of the constructor.

    |

    **Render Methods**

    :py:class:`KittyImage` provides two methods of :term:`rendering` images, namely:

    LINES (default)
       Renders an image line-by-line i.e the image is evenly split across the number
       of lines it should occupy.

       Pros:

       * Good for use cases where it might be required to trim some lines of the
         image.

    WHOLE
       Renders an image all at once i.e the entire image data is encoded into one
       line of the :term:`rendered` output, such that the entire image is drawn once
       by the terminal and still occupies the correct amount of lines and columns.

       Pros:

       * Render results are more compact (i.e less in character count) than with
         the **LINES** method since the entire image is encoded at once.

    The render method can be set with
    :py:meth:`set_render_method() <BaseImage.set_render_method>` using the names
    specified above.

    |

    **Style-Specific Render Parameters**

    See :py:meth:`BaseImage.draw` (particularly the *style* parameter).

    * **method** (*None | str*) → Render method override.

      * ``None`` → the current effective render method of the instance is used
      * A valid render method name (as specified in the **Render Methods** section
        above) → used instead of the current effective render method of the instance
      * *default* → ``None``

    * **z_index** (*int*) → The stacking order of graphics and text for
      **non-animations**.

      * An integer in the **signed 32-bit** range (excluding ``-(2**31)``)
      * ``>= 0`` → the image will be drawn above text
      * ``< 0`` → the image will be drawn below text
      * ``< -(2**31)/2`` → the image will be drawn below cells with non-default
        background color
      * *default* → ``0``
      * Overlapping graphics on different z-indexes will be blended (by the terminal
        emulator) if they are semi-transparent.
      * To inter-mix text with graphics, see the *mix* parameter.

    * **mix** (*bool*) → Graphics/Text inter-mix policy.

      * ``False`` → text within the region covered by the drawn render output will be
        erased, though text can be inter-mixed with graphics after drawing
      * ``True`` → text within the region covered by the drawn render output will NOT
        be erased
      * *default* → ``False``

    * **compress** (*int*) → ZLIB compression level.

      * ``0`` <= *compress* <= ``9``
      * ``1`` → best speed, ``9`` → best compression, ``0`` → no compression
      * *default* → ``4``
      * Results in a trade-off between render time and data size/draw speed

    |

    **Format Specification**

    See :ref:`format-spec`.

    ::

        [ <method> ]  [ z <z-index> ]  [ m <mix> ]  [ c <compress> ]

    * ``method`` → render method override

      * ``L`` → **LINES** render method (current frame only, for animated images)
      * ``W`` → **WHOLE** render method (current frame only, for animated images)
      * *default* → Current effective render method of the image

    * ``z`` → graphics/text stacking order

      * ``z-index`` → z-index

        * An integer in the **signed 32-bit** range (excluding ``-(2**31)``)
        * ``>= 0`` → the render output will be drawn above text
        * ``< 0`` → the render output will be drawn below text
        * ``< -(2**31)/2`` → the render output will be drawn below cells with
          non-default background color

      * *default* → ``z0`` (z-index zero)
      * e.g ``z0``, ``z1``, ``z-1``, ``z2147483647``, ``z-2147483648``
      * overlapping graphics on different z-indexes will be blended
        (by the terminal emulator) if they are semi-transparent

    * ``m`` → graphics/text inter-mix policy

      * ``mix`` → inter-mix policy

        * ``0`` → text within the region covered by the drawn render output will be
          erased, though text can be inter-mixed with graphics after drawing
        * ``1`` → text within the region covered by the drawn render output will NOT
          be erased

      * *default* → ``m0``
      * e.g ``m0``, ``m1``

    * ``c`` → ZLIB compression level

      * ``compress`` → compression level

        * An integer in the range ``0`` <= ``compress`` <= ``9``
        * ``1`` → best speed, ``9`` → best compression, ``0`` → no compression

      * *default* → ``c4``
      * e.g ``c0``, ``c9``
      * results in a trade-off between render time and data size/draw speed

    |

    IMPORTANT:
        Currently supported terminal emulators are:

        * `Kitty <https://sw.kovidgoyal.net/kitty/>`_ >= 0.20.0.
        * `Konsole <https://konsole.kde.org>`_ >= 22.04.0.
    """

    _FORMAT_SPEC: Tuple[re.Pattern] = tuple(
        map(re.compile, r"[LW] z-?\d+ m[01] c[0-9]".split(" "))
    )
    _render_methods: Set[str] = {LINES, WHOLE}
    _default_render_method: str = LINES
    _render_method: str = LINES
    _style_args = {
        "method": (
            None,
            (
                lambda x: isinstance(x, str),
                "Render method must be a string",
            ),
            (
                lambda x: x.lower() in __class__._render_methods,
                "Unknown render method for 'kitty' render style",
            ),
        ),
        "z_index": (
            0,
            (
                lambda x: isinstance(x, int),
                "z-index must be an integer",
            ),
            (
                # INT32_MIN is reserved for non-native animations
                lambda x: -(2**31) < x < 2**31,
                "z-index must be within the 32-bit signed integer range "
                "(excluding ``-(2**31)``)",
            ),
        ),
        "mix": (
            False,
            (
                lambda x: isinstance(x, bool),
                "Inter-mix policy must be a boolean",
            ),
            (lambda _: True, ""),
        ),
        "compress": (
            4,
            (
                lambda x: isinstance(x, int),
                "Compression level must be an integer",
            ),
            (
                lambda x: 0 <= x <= 9,
                "Compression level must be between 0 and 9, both inclusive",
            ),
        ),
    }

    _TERM: str = ""
    _TERM_VERSION: str = ""
    _KITTY_VERSION: Tuple[int, int, int] = ()

    @classmethod
    def clear(
        cls, *, cursor: bool = False, z_index: Optional[int] = None, now: bool = False
    ) -> None:
        """Clears images.

        Args:
            cursor: If ``True``, all images intersecting with the current cursor
              position are cleared.
            z_index: An integer in the **signed 32-bit range**. If given, all images
              on the given z-index are cleared.
            now: If ``True`` the images are cleared immediately, without affecting
              any standard I/O stream.
              Otherwise they're cleared when next :py:data:`sys.stdout` is flushed.

        Aside *now*, **only one** other argument may be given. If no argument is given
        (aside *now*) or default values are given, all images visible on the screen are
        cleared.

        NOTE:
            This method does nothing if the render style is not supported.
        """
        if not (cls._forced_support or cls.is_supported()):
            return

        if not isinstance(cursor, bool):
            raise arg_type_error("cursor", cursor)

        if z_index is not None:
            if not isinstance(z_index, int):
                raise arg_type_error("z_index", z_index)
            if not -(1 << 31) <= z_index < (1 << 31):
                raise arg_value_error_range("z_index", z_index)

        if not isinstance(now, bool):
            raise arg_type_error("now", now)

        default_args = __class__.clear.__func__.__kwdefaults__
        nonlocals = locals()
        args = {name: nonlocals[name] for name in default_args}
        given_args = args.items() - (default_args.items() | {("now", True)})

        if len(given_args) > 1:
            raise arg_value_error_msg(
                "Only one argument (aside 'now') may be given", len(given_args)
            )
        elif given_args:
            arg, _ = given_args.pop()
            (write_tty if now else _stdout_write)(
                (ctlseqs.KITTY_DELETE_CURSOR_b if now else ctlseqs.KITTY_DELETE_CURSOR)
                if arg == "cursor"
                else (
                    (
                        ctlseqs.KITTY_DELETE_Z_INDEX_b
                        if now
                        else ctlseqs.KITTY_DELETE_Z_INDEX
                    )
                    % z_index
                )
            )
        elif now:
            write_tty(ctlseqs.KITTY_DELETE_ALL_b)
        else:
            _stdout_write(ctlseqs.KITTY_DELETE_ALL)

    @classmethod
    def is_supported(cls) -> bool:
        if cls._supported is None:
            cls._supported = False

            # The graphics query for support detection messes up iTerm2's window title
            if get_terminal_name_version()[0] == "iterm2":
                return False

            # Kitty graphics query + terminal attribute query
            # The second query is to speed up the query since most (if not all)
            # terminals should support it and most terminals treat queries as FIFO
            response = query_terminal(
                ctlseqs.KITTY_SUPPORT_QUERY_b + ctlseqs.DA1_b,
                lambda s: not s.endswith(b"c"),
            )

            # Not supported if it doesn't respond to either query
            # or responds to the second but not the first
            if response:
                response = ctlseqs.KITTY_RESPONSE_re.match(response.decode())
            if response and response["id"] == "31" and response["message"] == "OK":
                name, version = get_terminal_name_version()
                # Only kitty >= 0.20.0 implement the protocol features utilized
                if name == "kitty" and version:
                    try:
                        version_tuple = tuple(map(int, version.split(".")))
                    except ValueError:  # Version string not "understood"
                        pass
                    else:
                        if version_tuple >= (0, 20, 0):
                            cls._TERM, cls._TERM_VERSION = name, version
                            cls._KITTY_VERSION = version_tuple
                            cls._supported = True
                # Konsole is good as long as it responds to the graphics query
                elif name == "konsole":
                    cls._TERM, cls._TERM_VERSION = name, version or ""
                    cls._supported = True

        return cls._supported

    @classmethod
    def _check_style_format_spec(cls, spec: str, original: str) -> Dict[str, Any]:
        parent, (method, z_index, mix, compress) = cls._get_style_format_spec(
            spec, original
        )
        args = {}
        if parent:
            args.update(super()._check_style_format_spec(parent, original))
        if method:
            args["method"] = LINES if method == "L" else WHOLE
        if z_index:
            args["z_index"] = int(z_index[1:])
        if mix:
            args["mix"] = bool(int(mix[-1]))
        if compress:
            args["compress"] = int(compress[-1])

        return cls._check_style_args(args)

    @classmethod
    def _clear_frame(cls) -> bool:
        """Clears an animation frame on-screen.

        | Only used on Kitty <= 0.25.0 because ``blend=False`` is buggy on these
          versions. Does nothing on any other version or terminal.
        | Also clears any frame of any previously drawn animation, since they all use
          the same z-index.

        See :py:meth:`~term_image.image.BaseImage._clear_frame` for description.
        """
        if cls._KITTY_VERSION and cls._KITTY_VERSION <= (0, 25, 0):
            cls.clear(z_index=-(1 << 31))
            return True
        return False

    def _display_animated(self, *args, **kwargs) -> None:
        kwargs["z_index"] = -(1 << 31)
        if self._KITTY_VERSION > (0, 25, 0):
            kwargs["blend"] = False

        super()._display_animated(*args, **kwargs)

    @staticmethod
    def _handle_interrupted_draw():
        """Performs necessary actions when image drawing is interrupted.

        If drawing is interrupted while transmitting a command, it causes terminal to
        wait for more data (in fact, it actually consumes any output following)
        until the output reaches the expected payload size or ST (String Terminator)
        is written.

        Also, if the image data was chunked, it would be expecting the last chunk.
        In this case, output is not consumed but the next graphics command sent
        might not be treated as expected on some terminals e.g Konsole.
        """

        # End last command (does no harm if there wasn't an unterminated command)
        # and send "last chunk" in case the last transmission was chunked.
        # Konsole sometimes requires ST to be written twice.
        print(ctlseqs.ST * 2 + ctlseqs.KITTY_END_CHUNKED, end="", flush=True)

    def _render_image(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        frame: bool = False,
        method: Optional[str] = None,
        z_index: int = 0,
        mix: bool = False,
        compress: int = 4,
        blend: bool = True,
    ) -> str:
        """See :py:meth:`BaseImage._render_image` for the description of the method and
        :py:meth:`draw` for parameters not described here.

        Args:
            blend: If ``False``, the rendered image deletes overlapping/intersecting
              images when drawn. Otherwise, the behaviour is dependent on the z-index
              and/or the terminal emulator (for images with the same z-index).
        """
        # NOTE: It's more efficient to write separate strings to the buffer separately
        # than concatenate and write together.

        # Using `c` and `r` ensures that an image always occupies the correct amount
        # of columns and lines even if the cell size has changed when it's drawn.
        # Since we use `c` and `r` control data keys, there's no need upscaling the
        # image on this end to reduce payload.
        # Anyways, this also implies that the image(s) have to be resized by the
        # terminal emulator, thereby leaving various details of resizing in the hands
        # of the terminal emulator such as the resampling method, etc.
        # This particularly affects the LINES render method negatively, resulting in
        # slant/curved edges not lining up across lines (amongst other artifacts
        # observed on Konsole) supposedly because the terminal emulator resizes each
        # line separately.
        # Hence, this optimization is only used for the WHOLE render method.

        render_method = (method or self._render_method).lower()
        r_width, r_height = self.rendered_size
        width, height = (
            self._get_minimal_render_size()
            if render_method == WHOLE
            else self._get_render_size()
        )

        frame_img = img if frame else None
        img = self._get_render_data(
            img, alpha, size=(width, height), pixel_data=False, frame=frame  # fmt: skip
        )[0]
        format = getattr(f, img.mode)
        raw_image = img.tobytes()

        # clean up (ImageIterator uses one PIL image throughout)
        if frame_img is not img:
            self._close_image(img)

        control_data = ControlData(f=format, s=width, c=r_width, z=z_index)
        fill = ("" if mix else ERASE_CHARS % r_width) + (CURSOR_FORWARD % r_width)
        fill_newline = fill + "\n"

        if render_method == LINES:
            cell_height = height // r_height
            bytes_per_line = width * cell_height * (format // 8)
            vars(control_data).update(dict(v=cell_height, r=1))

            with io.StringIO() as buffer, io.BytesIO(raw_image) as raw_image:
                trans = Transmission(
                    control_data, raw_image.read(bytes_per_line), compress
                )
                blend or buffer.write(KITTY_DELETE_CURSOR)
                for chunk in trans.get_chunks():
                    buffer.write(chunk)
                for _ in range(r_height - 1):
                    buffer.write(fill_newline)
                    trans = Transmission(
                        control_data, raw_image.read(bytes_per_line), compress
                    )
                    blend or buffer.write(KITTY_DELETE_CURSOR)
                    for chunk in trans.get_chunks():
                        buffer.write(chunk)
                buffer.write(fill)

                return buffer.getvalue()

        vars(control_data).update(v=height, r=r_height)
        return "".join(
            (
                KITTY_DELETE_CURSOR * (not blend),
                Transmission(control_data, raw_image, compress).get_chunked(),
                fill_newline * (r_height - 1),
                fill,
            )
        )


@dataclass
class Transmission:
    """An abstraction of the kitty terminal graphics escape code.

    Args:
        control: The control data.
        payload: The payload.
        level: Compression level.
    """

    control: ControlData
    payload: bytes

    # From tests with a few images, from anything beyond 4, the decrease in size
    # doesn't seem to be worth the increase in compression time in most cases.
    # Might change if proven otherwise.
    level: int = 4

    def __post_init__(self):
        self._compressed = False
        if self.level:
            self.compress()
        else:
            self.control.o = None

    def compress(self):
        if self.control.t == t.DIRECT and not self._compressed and self.level:
            self.payload = compress(self.payload, self.level)
            self.control.o = o.ZLIB
            self._compressed = True

    def decompress(self):
        if self.control.t == t.DIRECT and self._compressed:
            self.control.o = None
            self.payload = decompress(self.payload)
            self._compressed = False

    def encode(self) -> bytes:
        return standard_b64encode(self.payload)

    def get_chunked(self) -> str:
        return "".join(self.get_chunks())

    def get_chunks(self, size: int = 4096) -> Generator[str, None, None]:
        with self.get_payload() as payload:
            chunk, next_chunk = payload.read(size), payload.read(size)
            yield (
                KITTY_TRANSMISSION
                % (f"{self.get_control_data()},m={bool(next_chunk):d}", chunk)
            )

            chunk, next_chunk = next_chunk, payload.read(size)
            while next_chunk:
                yield KITTY_TRANSMISSION % ("m=1", chunk)
                chunk, next_chunk = next_chunk, payload.read(size)

            if chunk:  # false if there was never a next chunk
                yield KITTY_TRANSMISSION % ("m=0", chunk)

    def get_control_data(self) -> str:
        return ",".join(
            f"{key}={value}"
            for key, value in asdict(self.control).items()
            if value is not None
        )

    def get_payload(self) -> io.StringIO:
        return io.StringIO(self.encode().decode("ascii"))


# Values for control data keys with limited set of values


class a:
    TRANS = "t"
    TRANS_DISP = "T"
    QUERY = "q"
    PLACE = "p"
    DELETE = "d"
    TRANS_FRAMES = "f"
    CONTROL_ANIM = "a"
    COMPOSE_FRAMES = "c"


class C:
    MOVE = 0
    STAY = 1


class f:
    RGB = 24
    RGBA = 32
    PNG = 100


class o:
    ZLIB = "z"


class t:
    DIRECT = "d"
    FILE = "f"
    TEMP = "t"
    SHARED = "s"


class z:
    BEHIND = -1
    IN_FRONT = 0


@dataclass
class ControlData:
    """Represents a portion of the kitty terminal graphics protocol control data"""

    a: Optional[str] = a.TRANS_DISP  # action
    f: Optional[int] = f.RGBA  # data format
    t: Optional[str] = t.DIRECT  # transmission medium
    s: Optional[int] = None  # image width
    v: Optional[int] = None  # image height
    z: Optional[int] = z.IN_FRONT  # z-index
    o: Optional[str] = None  # compression
    C: Optional[int] = C.STAY  # cursor movement policy

    # # Image display size in columns and rows/lines
    # # The image is shrunk or enlarged to fit
    c: Optional[int] = None  # columns
    r: Optional[int] = None  # rows

    def __post_init__(self):
        if self.f == f.PNG:
            self.s = self.v = None


class _ControlData:  # Currently Unused
    i: Optional[int] = None  # image ID
    d: Optional[str] = None  # delete images
    m: Optional[int] = None  # payload chunk
    O: Optional[int] = None  # data start offset; with t=s or t=f
    S: Optional[int] = None  # data size in bytes; with f=100,o=z or t=s or t=f

    # Origin offset (px) within the current cell; Must be less than the cell size
    # (0, 0) == Top-left corner of the cell; Not used with `c` and `r`.
    X: Optional[int] = None
    Y: Optional[int] = None

    # Image crop (px)
    # # crop origin; (0, 0) == top-left corner of the image
    x: Optional[int] = None
    y: Optional[int] = None
    # # crop rectangle size
    w: Optional[int] = None
    h: Optional[int] = None


_stdout_write = sys.stdout.write



"""
..
   Control Sequences

   See https://invisible-island.net/xterm/ctlseqs/ctlseqs.html
"""

from __future__ import annotations

__all__ = []  # Updated later on

import re
from typing import Tuple

# Parameters
C = "%c"
Ps = "%d"
Pt = "%s"
Pm = lambda n: ";".join((Ps,) * n)  # noqa: E731

_START = None  # Marks the beginning control sequence definitions

# C0
BEL = "\x07"
ESC = "\x1b"

# C1
APC = f"{ESC}_"
CSI = f"{ESC}["
DCS = f"{ESC}P"
OSC = f"{ESC}]"
ST = f"{ESC}\\"

# Cursor Movement
CURSOR_UP = f"{CSI}{Ps}A"
CURSOR_DOWN = f"{CSI}{Ps}B"
CURSOR_FORWARD = f"{CSI}{Ps}C"
CURSOR_BACKWARD = f"{CSI}{Ps}D"

# Select Graphic Rendition
SGR_NORMAL = f"{CSI}m"
SGR_FG_RGB = f"{CSI}38;2;{Pm(3)}m"
SGR_FG_RGB_2 = f"{CSI}38:2::{Ps}:{Ps}:{Ps}m"
SGR_BG_RGB = f"{CSI}48;2;{Pm(3)}m"
SGR_BG_RGB_2 = f"{CSI}48:2::{Ps}:{Ps}:{Ps}m"
SGR_FG_INDEXED = f"{CSI}38;5;{Ps}m"
SGR_FG_INDEXED_2 = f"{CSI}38:5:{Ps}m"
SGR_BG_INDEXED = f"{CSI}48;5;{Ps}m"
SGR_BG_INDEXED_2 = f"{CSI}48:5:{Ps}m"

# DEC Modes
DECSET = f"{CSI}?{Ps}h"
DECRST = f"{CSI}?{Ps}l"

SHOW_CURSOR = DECSET % 25
HIDE_CURSOR = DECRST % 25

# # Terminal Synchronized Output
# # See https://gist.github.com/christianparpart/d8a62cc1ab659194337d73e399004036
BEGIN_SYNCED_UPDATE = DECSET % 2026
END_SYNCED_UPDATE = DECRST % 2026

# Window manipulation (XTWINOPS)
XTWINOPS_1 = f"{CSI}{Ps}t"

TEXT_AREA_SIZE_PX = XTWINOPS_1 % 14
CELL_SIZE_PX = XTWINOPS_1 % 16

# Text parameters (OSC Ps ; Pt ST)
TEXT_PARAM_SET = f"{OSC}{Ps};{Pt}{ST}"
TEXT_PARAM_QUERY = f"{OSC}{Ps};?{ST}"

TEXT_FG_QUERY = TEXT_PARAM_QUERY % 10
TEXT_BG_QUERY = TEXT_PARAM_QUERY % 11

# Others
DA1 = f"{CSI}c"
ERASE_CHARS = f"{CSI}{Ps}X"
XTVERSION = f"{CSI}>q"

# iTerm2 Inline Image Protocol
# See https://iterm2.com/documentation-images.html
ITERM2_START = f"{OSC}1337;File="

# Kitty Graphics Protocol
# See https://sw.kovidgoyal.net/kitty/graphics-protocol/
KITTY_START = f"{APC}G"
KITTY_TRANSMISSION = f"{KITTY_START}{Pt};{Pt}{ST}"
KITTY_DELETE = KITTY_TRANSMISSION % (f"a=d,d={C}", "")
KITTY_DELETE_EXTRA = KITTY_TRANSMISSION % (f"a=d,d={C},{Pt}", "")

KITTY_SUPPORT_QUERY = KITTY_TRANSMISSION % (
    "a=q,t=d,i=31,f=24,s=1,v=1,C=1,c=1,r=1",
    "AAAA",
)
KITTY_END_CHUNKED = KITTY_TRANSMISSION % ("q=1,m=0", "")
KITTY_DELETE_ALL = KITTY_DELETE % "A"
KITTY_DELETE_CURSOR = KITTY_DELETE % "C"
KITTY_DELETE_Z_INDEX = KITTY_DELETE_EXTRA % ("Z", f"z={Ps}")


module_items = tuple(globals().items())
for name, value in module_items[module_items.index(("_START", None)) + 1 :]:
    globals()[f"{name}_b"] = value.encode()
    __all__.extend((name, f"{name}_b"))


# Patterns for responses to queries
class Response:
    CSI_escaped = re.escape(CSI)
    OSC_escaped = re.escape(OSC)
    ST_escaped = re.escape(ST)
    ST_or_BEL = f"(?:{ST_escaped}|{BEL})"
    XTWINOPS = rf"{CSI_escaped}{Ps};(\d+);(\d+)t"

    RGB_SPEC_re = rf"{OSC_escaped}(\d+);rgb:([\da-fA-F/]+){ST_or_BEL}"
    XTVERSION_re = rf"{DCS}>\|(\w+)[( ]([^){ESC}]+)\)?{ST_or_BEL}"
    TEXT_AREA_SIZE_PX_re = XTWINOPS % 4
    CELL_SIZE_PX_re = XTWINOPS % 6
    KITTY_RESPONSE_re = (
        f"{KITTY_START}"
        r"i=(?P<id>\d+)(?:,I=(?P<number>\d+))?;(?P<message>.+?)"
        f"{ST_escaped}"
    )

    for name, regex in tuple(locals().items()):
        if name.endswith("_re"):
            globals()[name] = re.compile(regex, re.ASCII)
            __all__.append(name)


__all__ += ("x_parse_color",)


def x_parse_color(spec: str) -> Tuple[int, int, int]:
    """Converts an RGB device specification according to ``XParseColor``

    See the "Color Names" section of the ``XParseColor`` man page.
    """
    # One hex char -> 4 bits
    return tuple(int(x, 16) * 255 // ((1 << (len(x) * 4)) - 1) for x in spec.split("/"))


del _START, module_items, Response



from __future__ import annotations

__all__ = ("BlockImage",)

import io
import os
import re
from math import ceil
from operator import mul
from typing import Any, Optional, Tuple, Union

import PIL

from ..ctlseqs import SGR_BG_INDEXED, SGR_BG_RGB, SGR_FG_INDEXED, SGR_FG_RGB, SGR_NORMAL
from ..utils import get_fg_bg_colors
from .common import TextImage

LOWER_PIXEL = "\u2584"  # lower-half block element
UPPER_PIXEL = "\u2580"  # upper-half block element

# Constants for render methods
DIRECT = "direct"
INDEXED = "indexed"


class BlockImage(TextImage):
    """A render style using Unicode half blocks with direct-color or indexed-color
    control sequences.

    See :py:class:`TextImage` for the description of the constructor.

    |

    **Render Methods**

    :py:class:`BlockImage` provides two methods of :term:`rendering` images, namely:

    DIRECT (default)
       Renders an image using direct-color (truecolor) control sequences.

       Pros:

       * Better color reproduction.

       Cons:

       * Lesser terminal emulator support (though any terminal emulator worthy of use
         today actually does provide support).

    INDEXED
       Renders an image using indexed-color control sequences but with only the upper
       **240 colors** of the terminal's 256-color palette.

       Pros:

       * Wider terminal emulator support.

       Cons:

       * Worse color reproduction.

    The render method can be set with
    :py:meth:`set_render_method() <BaseImage.set_render_method>` using the names
    specified above.

    |

    **Style-Specific Render Parameters**

    See :py:meth:`BaseImage.draw` (particularly the *style* parameter).

    * **method** (*None | str*) → Render method override.

      * ``None`` → the current effective render method of the instance is used
      * A valid render method name (as specified in the **Render Methods** section
        above) → used instead of the current effective render method of the instance
      * *default* → ``None``

    |

    **Format Specification**

    See :ref:`format-spec`.

    ::

        [ <method> ]

    * ``method`` → render method override

      * ``D`` → **DIRECT** render method (current frame only, for animated images)
      * ``I`` → **INDEXED** render method (current frame only, for animated images)
      * *default* → Current effective render method of the image
    """

    _FORMAT_SPEC: tuple[re.Pattern] = (re.compile("[DI]"),)
    _render_methods: set[str] = {DIRECT, INDEXED}
    _default_render_method: str = DIRECT
    _render_method: str = DIRECT
    _style_args = {
        "method": (
            None,
            (
                lambda x: isinstance(x, str),
                "Render method must be a string",
            ),
            (
                lambda x: x.lower() in __class__._render_methods,
                "Unknown render method for 'block' render style",
            ),
        ),
    }

    @classmethod
    def is_supported(cls):
        if cls._supported is None:
            COLORTERM = os.environ.get("COLORTERM") or ""
            TERM = os.environ.get("TERM") or ""
            cls._supported = (
                "truecolor" in COLORTERM or "24bit" in COLORTERM or "256color" in TERM
            )

        return cls._supported

    @classmethod
    def _check_style_format_spec(cls, spec: str, original: str) -> dict[str, Any]:
        parent, (method,) = cls._get_style_format_spec(spec, original)
        args = {}
        if parent:
            args.update(super()._check_style_format_spec(parent, original))
        if method:
            args["method"] = DIRECT if method == "D" else INDEXED

        return cls._check_style_args(args)

    def _get_render_size(self) -> Tuple[int, int]:
        return tuple(map(mul, self.rendered_size, (1, 2)))

    @staticmethod
    def _pixels_cols(
        *, pixels: Optional[int] = None, cols: Optional[int] = None
    ) -> int:
        return pixels if pixels is not None else cols

    @staticmethod
    def _pixels_lines(
        *, pixels: Optional[int] = None, lines: Optional[int] = None
    ) -> int:
        return ceil(pixels / 2) if pixels is not None else lines * 2

    def _render_image(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        frame: bool = False,
        method: str | None = None,
        split_cells: bool = False,
    ) -> str:
        # NOTE:
        # It's more efficient to write separate strings to the buffer separately
        # than concatenate and write together.

        def update_buffer():
            if alpha:
                no_alpha = False
                if a_cluster1 == 0 == a_cluster2:
                    buf_write(SGR_NORMAL)
                    buf_write(blank * n)
                elif a_cluster1 == 0:  # up is transparent
                    buf_write(SGR_NORMAL)
                    buf_write(sgr_fg % cluster2)
                    buf_write(lower_pixel * n)
                elif a_cluster2 == 0:  # down is transparent
                    buf_write(SGR_NORMAL)
                    buf_write(sgr_fg % cluster1)
                    buf_write(upper_pixel * n)
                else:
                    no_alpha = True

            if not alpha or no_alpha:
                if method_is_direct:
                    r, g, b = cluster2
                    # Kitty does not render BG colors equal to the default BG color
                    if is_on_kitty and cluster2 == bg_color:
                        r += r < 255 or -1
                    buf_write(sgr_bg % (r, g, b))
                else:
                    buf_write(sgr_bg % cluster2)

                if cluster1 == cluster2:
                    buf_write(blank * n)
                else:
                    buf_write(sgr_fg % cluster1)
                    buf_write(upper_pixel * n)

        buffer = io.StringIO()
        buf_write = buffer.write  # Eliminate attribute resolution cost

        render_method = (method or self._render_method).lower()
        method_is_direct = render_method == DIRECT
        if method_is_direct:
            bg_color = get_fg_bg_colors()[1]
            is_on_kitty = self._is_on_kitty()
        if split_cells:
            blank = " \0"
            lower_pixel = LOWER_PIXEL + "\0"
            upper_pixel = UPPER_PIXEL + "\0"
        else:
            blank = " "
            lower_pixel = LOWER_PIXEL
            upper_pixel = UPPER_PIXEL
        end_of_line = SGR_NORMAL + "\n"
        sgr_fg = SGR_FG_RGB if method_is_direct else SGR_FG_INDEXED
        sgr_bg = SGR_BG_RGB if method_is_direct else SGR_BG_INDEXED

        width, height = self._get_render_size()
        frame_img = img if frame else None
        img, rgb, a = self._get_render_data(
            img,
            alpha,
            round_alpha=True,
            frame=frame,
            indexed_color=not method_is_direct,
        )
        alpha = img.mode == ("RGBA" if method_is_direct else "PA")

        # clean up (ImageIterator uses one PIL image throughout)
        if frame_img is not img:
            self._close_image(img)

        rgb_pairs = (
            (
                zip(rgb[x : x + width], rgb[x + width : x + width * 2]),
                (rgb[x], rgb[x + width]),
            )
            for x in range(0, len(rgb), width * 2)
        )
        a_pairs = (
            (
                zip(a[x : x + width], a[x + width : x + width * 2]),
                (a[x], a[x + width]),
            )
            for x in range(0, len(a), width * 2)
        )

        row_no = 0
        # Two rows of pixels per line
        for (rgb_pair, (cluster1, cluster2)), (a_pair, (a_cluster1, a_cluster2)) in zip(
            rgb_pairs, a_pairs
        ):
            row_no += 2
            n = 0
            for (px1, px2), (a1, a2) in zip(rgb_pair, a_pair):
                # Color-code characters and write to buffer
                # when upper and/or lower pixel color/alpha-level changes
                if not (alpha and a1 == a_cluster1 == 0 == a_cluster2 == a2) and (
                    px1 != cluster1
                    or px2 != cluster2
                    or alpha
                    and (
                        # From non-transparent to transparent
                        a_cluster1 != a1 == 0
                        or a_cluster2 != a2 == 0
                        # From transparent to non-transparent
                        or 0 == a_cluster1 != a1
                        or 0 == a_cluster2 != a2
                    )
                ):
                    update_buffer()
                    cluster1 = px1
                    cluster2 = px2
                    if alpha:
                        a_cluster1 = a1
                        a_cluster2 = a2
                    n = 0
                n += 1

            update_buffer()  # Rest of the line
            if split_cells:
                # Set the last "\0" to be overwritten by the next byte
                buffer.seek(buffer.tell() - 1)
            if row_no < height:  # last line not yet rendered
                buf_write(end_of_line)

        buf_write(SGR_NORMAL)  # Reset color after last line

        with buffer:
            return buffer.getvalue()



"""
term-image

Display images in the terminal

Copyright (c) 2022, Toluwaleke Ogundipe <anonymoux47@gmail.com>
"""

from __future__ import annotations

__all__ = (
    "DEFAULT_QUERY_TIMEOUT",
    "AutoCellRatio",
    "disable_queries",
    "disable_win_size_swap",
    "enable_queries",
    "enable_win_size_swap",
    "get_cell_ratio",
    "set_cell_ratio",
    "set_query_timeout",
)
__author__ = "Toluwaleke Ogundipe"

from enum import Enum, auto
from operator import truediv
from typing import Optional, Union

from . import utils
from .exceptions import TermImageError

version_info = (0, 7, 2)

# Follows https://semver.org/spec/v2.0.0.html
__version__ = ".".join(map(str, version_info[:3]))
if version_info[3:]:
    __version__ += "-" + ".".join(map(str, version_info[3:]))

#: Default timeout for :ref:`terminal-queries`
#:
#: See also: :py:func:`set_query_timeout`
DEFAULT_QUERY_TIMEOUT: float = utils._query_timeout  # Final[float]


class AutoCellRatio(Enum):
    """Values for setting :ref:`auto-cell-ratio`."""

    is_supported: Optional[bool]

    FIXED = auto()
    DYNAMIC = auto()


def disable_queries() -> None:
    """Disables :ref:`terminal-queries`.

    To re-enable queries, call :py:func:`enable_queries`.

    NOTE:
        This affects all :ref:`dependent features <queried-features>`.
    """
    utils._queries_enabled = False


def disable_win_size_swap():
    """Disables a workaround for terminal emulators that wrongly report window
    dimensions swapped.

    This workaround is disabled by default. While disabled, the window dimensions
    reported by the :term:`active terminal` are used as-is.

    NOTE:
        This affects :ref:`auto-cell-ratio` computation and size computations for
        :ref:`graphics-based`.
    """
    if utils._swap_win_size:
        utils._swap_win_size = False
        with utils._cell_size_lock:
            utils._cell_size_cache[:] = (0,) * 4


def enable_queries() -> None:
    """Re-Enables :ref:`terminal-queries`.

    Queries are enabled by default. To disable, call :py:func:`disable_queries`.

    NOTE:
        This affects all :ref:`dependent features <queried-features>`.
    """
    if not utils._queries_enabled:
        utils._queries_enabled = True
        utils.get_fg_bg_colors._invalidate_cache()
        utils.get_terminal_name_version._invalidate_cache()
        with utils._cell_size_lock:
            utils._cell_size_cache[:] = (0,) * 4


def enable_win_size_swap():
    """Enables a workaround for terminal emulators that wrongly report window
    dimensions swapped.

    While enabled, the window dimensions reported by the :term:`active terminal` are
    swapped. This workaround is required on some older VTE-based terminal emulators.

    NOTE:
        This affects :ref:`auto-cell-ratio` computation and size computations for
        :ref:`graphics-based`.
    """
    if not utils._swap_win_size:
        utils._swap_win_size = True
        with utils._cell_size_lock:
            utils._cell_size_cache[:] = (0,) * 4


def get_cell_ratio() -> float:
    """Returns the global :term:`cell ratio`.

    See :py:func:`set_cell_ratio`.
    """
    # `(1, 2)` is a fallback in case the terminal doesn't respond in time
    return _cell_ratio or truediv(*(utils.get_cell_size() or (1, 2)))


def set_cell_ratio(ratio: Union[float, AutoCellRatio]) -> None:
    """Sets the global :term:`cell ratio`.

    Args:
        ratio: Can be one of the following values.

          * A positive :py:class:`float` value.
          * :py:attr:`AutoCellRatio.FIXED`, the ratio is immediately determined from
            the :term:`active terminal`.
          * :py:attr:`AutoCellRatio.DYNAMIC`, the ratio is determined from the
            :term:`active terminal` whenever :py:func:`get_cell_ratio` is called,
            though with some caching involved, such that the ratio is re-determined
            only if the terminal size changes.

    Raises:
        TypeError: An argument is of an inappropriate type.
        ValueError: An argument is of an appropriate type but has an
          unexpected/invalid value.
        term_image.exceptions.TermImageError: Auto cell ratio is not supported
          in the :term:`active terminal` or on the current platform.

    This value is taken into consideration when setting image sizes for **text-based**
    render styles, in order to preserve the aspect ratio of images drawn to the
    terminal.

    NOTE:
        Changing the cell ratio does not automatically affect any image that has a
        :term:`fixed size`. For a change in cell ratio to take effect, the image's
        size has to be re-set.

    ATTENTION:
        See :ref:`auto-cell-ratio` for details about the auto modes.
    """
    global _cell_ratio

    if isinstance(ratio, AutoCellRatio):
        if AutoCellRatio.is_supported is None:
            AutoCellRatio.is_supported = utils.get_cell_size() is not None

        if not AutoCellRatio.is_supported:
            raise TermImageError(
                "Auto cell ratio is not supported in the active terminal or on the "
                "current platform"
            )
        elif ratio is AutoCellRatio.FIXED:
            # `(1, 2)` is a fallback in case the terminal doesn't respond in time
            _cell_ratio = truediv(*(utils.get_cell_size() or (1, 2)))
        else:
            _cell_ratio = None
    elif isinstance(ratio, float):
        if ratio <= 0.0:
            raise utils.arg_value_error_range("ratio", ratio)
        _cell_ratio = ratio
    else:
        raise utils.arg_type_error("ratio", ratio)


def set_query_timeout(timeout: float) -> None:
    """Sets the timeout for :ref:`terminal-queries`.

    Args:
        timeout: Time limit for awaiting a response from the terminal, in seconds.

    Raises:
        TypeError: *timeout* is not a float.
        ValueError: *timeout* is less than or equal to zero.
    """
    if not isinstance(timeout, float):
        raise utils.arg_type_error("timeout", timeout)
    if timeout <= 0.0:
        raise utils.arg_value_error_range("timeout", timeout)

    utils._query_timeout = timeout


_cell_ratio = 0.5
AutoCellRatio.is_supported = None



from __future__ import annotations

__all__ = ("ITerm2Image",)

import io
import os
import re
import sys
import warnings
from base64 import standard_b64encode
from operator import mul
from typing import Any, Dict, Optional, Set, Tuple, Union

import PIL

from .. import ctlseqs

# These sequences are used during performance-critical operations that occur often
from ..ctlseqs import CURSOR_FORWARD, CURSOR_UP, ERASE_CHARS, ITERM2_START, ST
from ..exceptions import RenderError, TermImageWarning
from ..utils import (
    ClassInstanceProperty,
    ClassProperty,
    arg_type_error,
    arg_value_error_range,
    get_terminal_name_version,
    write_tty,
)
from .common import GraphicsImage, ImageMeta, ImageSource

# Constants for render methods
LINES = "lines"
WHOLE = "whole"
ANIM = "anim"


class ITerm2ImageMeta(ImageMeta):
    """Type of iterm2 render style classes."""

    __native_anim_max_bytes = _native_anim_max_bytes = 2 * 2**20  # 2 MiB default

    jpeg_quality = ClassInstanceProperty(
        lambda self: getattr(self, "_jpeg_quality", -1),
        doc="""JPEG encoding quality

        See the base instance of this metaclass for the complete description.
        """,
    )

    @jpeg_quality.setter
    def jpeg_quality(self, quality: int) -> None:
        if not isinstance(quality, int):
            raise arg_type_error("jpeg_quality", quality)
        if quality > 95:
            raise arg_value_error_range("jpeg_quality", quality)

        self._jpeg_quality = quality

    @jpeg_quality.deleter
    def jpeg_quality(self) -> None:
        try:
            del self._jpeg_quality
        except AttributeError:
            pass

    native_anim_max_bytes = ClassProperty(
        lambda self: __class__._native_anim_max_bytes,
        doc="""Maximum size (in bytes) of image data for native animation

        See the base instance of this metaclass for the complete description.
        """,
    )

    @native_anim_max_bytes.setter
    def native_anim_max_bytes(self, max_bytes: int):
        if not isinstance(max_bytes, int):
            raise arg_type_error("native_anim_max_bytes", max_bytes)
        if max_bytes <= 0:
            raise arg_value_error_range("native_anim_max_bytes", max_bytes)

        __class__._native_anim_max_bytes = max_bytes

    @native_anim_max_bytes.deleter
    def native_anim_max_bytes(self):
        __class__._native_anim_max_bytes = __class__.__native_anim_max_bytes

    read_from_file = ClassInstanceProperty(
        lambda self: getattr(self, "_read_from_file", True),
        doc="""Read-from-file optimization

        See the base instance of this metaclass for the complete description.
        """,
    )

    @read_from_file.setter
    def read_from_file(self, policy: bool) -> None:
        if not isinstance(policy, bool):
            raise arg_type_error("read_from_file", policy)

        self._read_from_file = policy

    @read_from_file.deleter
    def read_from_file(self) -> None:
        try:
            del self._read_from_file
        except AttributeError:
            pass


class ITerm2Image(GraphicsImage, metaclass=ITerm2ImageMeta):
    """A render style using the iTerm2 inline image protocol.

    See :py:class:`GraphicsImage` for the complete description of the constructor.

    |

    **Render Methods**

    :py:class:`ITerm2Image` provides two methods of :term:`rendering` images, namely:

    LINES (default)
       Renders an image line-by-line i.e the image is evenly split across the number
       of lines it should occupy.

       Pros:

       * Good for use cases where it might be required to trim some lines of the image.

       Cons:

       * Image drawing is significantly slower on iTerm2 due to the terminal emulator's
         performance.

    WHOLE
       Renders an image all at once i.e the entire image data is encoded into one
       line of the :term:`render` output, such that the entire image is drawn once
       by the terminal and still occupies the correct amount of lines and columns.

       Pros:

       * Render results are more compact (i.e less in character count) than with
         the **LINES** method since the entire image is encoded at once.
       * Image drawing is faster than with **LINES** on most terminals.
       * Smoother animations.

       Cons:

       * This method currently doesn't work well on iTerm2 and WezTerm when the image
         height is greater than the terminal height.

    ANIM
        Renders an animated image to utilize the protocol's native animation feature
        [1]_.

        Similar to the **WHOLE** render method, except that the terminal emulator
        animates the image, provided it supports the feature of the protocol.
        The animation is completely controlled by the terminal emulator.

        .. note::
            * If the image data size (in bytes) is greater than the value of
              :py:attr:`native_anim_max_bytes`, a warning is issued.
            * If used with :py:class:`~term_image.image.ImageIterator` or an animation,
              the **WHOLE** render method is used instead.
            * If the image is non-animated, the **WHOLE** render method is used instead.

    NOTE:
        The **LINES** method is the default only because it works properly in all cases,
        it's more advisable to use the **WHOLE** method except when the image height is
        greater than the terminal height or when trimming the image is required.

    The render method can be set with
    :py:meth:`set_render_method() <BaseImage.set_render_method>` using the names
    specified above.

    |

    **Style-Specific Render Parameters**

    See :py:meth:`BaseImage.draw` (particularly the *style* parameter).

    * **method** (*None | str*) → Render method override.

      * ``None`` → the current effective render method of the instance is used
      * A valid render method name (as specified in the **Render Methods** section
        above) → used instead of the current effective render method of the instance
      * *default* → ``None``

    * **mix** (*bool*) → Cell content inter-mix policy (**Only supported on WezTerm**,
      ignored otherwise).

      * ``False`` → existing contents of cells within the region covered by
        the drawn render output are erased
      * ``True`` → existing cell contents show under transparent areas of the
        drawn render output
      * *default* → ``False``

    * **compress** (*int*) → ZLIB compression level, for renders re-encoded in PNG
      format.

      * ``0`` <= *compress* <= ``9``
      * ``1`` → best speed, ``9`` → best compression, ``0`` → no compression
      * *default* → ``4``
      * Results in a trade-off between render time and data size/draw speed

    |

    **Format Specification**

    See :ref:`format-spec`.

    ::

        [ <method> ]  [ m <mix> ]  [ c <compress> ]

    * ``method`` → render method override

      * ``L`` → **LINES** render method (current frame only, for animated images)
      * ``W`` → **WHOLE** render method (current frame only, for animated images)
      * ``A`` → **ANIM** render method [1]_
      * *default* → current effective render method of the instance

    * ``m`` → cell content inter-mix policy (**Only supported in WezTerm**, ignored
      otherwise)

      * ``mix`` → inter-mix policy

        * ``0`` → existing contents of cells in the region covered by the drawn
          render output will be erased
        * ``1`` → existing cell contents show under transparent areas of the drawn
          render output

      * *default* → ``m0``
      * e.g ``m0``, ``m1``

    * ``c`` → ZLIB compression level, for renders re-encoded in PNG format

      * ``compress`` → compression level

        * An integer in the range ``0`` <= ``x`` <= ``9``
        * ``1`` → best speed, ``9`` → best compression, ``0`` → no compression

      * *default* → ``c4``
      * e.g ``c0``, ``c9``
      * Results in a trade-off between render time and data size/draw speed

    |

    IMPORTANT:
        Currently supported terminal emulators are:

        * `iTerm2 <https://iterm2.com>`_
        * `Konsole <https://konsole.kde.org>`_ >= 22.04.0
        * `WezTerm <https://wezfurlong.org/wezterm/>`_


    .. [1] Native animation support:

       * Not all animated image formats may be supported by every supported terminal
         emulator
       * Not all supported terminal emulators implement this feature of the protocol
         e.g on Konsole, the first frame is drawn but the image is not animated

    |
    """

    _FORMAT_SPEC: Tuple[re.Pattern] = tuple(
        map(re.compile, "[LWA] m[01] c[0-9]".split(" "))
    )
    _render_methods: Set[str] = {LINES, WHOLE, ANIM}
    _default_render_method: str = LINES
    _render_method: str = LINES
    _style_args = {
        "method": (
            None,
            (
                lambda x: isinstance(x, str),
                "Render method must be a string",
            ),
            (
                lambda x: x.lower() in __class__._render_methods,
                "Unknown render method for 'iterm2' render style",
            ),
        ),
        "mix": (
            False,
            (
                lambda x: isinstance(x, bool),
                "Cell content inter-mix policy must be a boolean",
            ),
            (lambda _: True, ""),
        ),
        "compress": (
            4,
            (
                lambda x: isinstance(x, int),
                "Compression level must be an integer",
            ),
            (
                lambda x: 0 <= x <= 9,
                "Compression level must be between 0 and 9, both inclusive",
            ),
        ),
    }

    _TERM: str = ""
    _TERM_VERSION: str = ""

    jpeg_quality = ClassInstanceProperty(
        ITerm2ImageMeta.jpeg_quality.fget,
        ITerm2ImageMeta.jpeg_quality.fset,
        ITerm2ImageMeta.jpeg_quality.fdel,
        doc="""JPEG encoding quality

        :type: int

        GET:
            Returns the effective JPEG encoding quality of the invoker
            (class or instance).

        SET:
            If invoked via:

            * a **class**, the **class-wide** quality is set.
            * an **instance**, the **instance-specific** quality is set.

        DELETE:
            If invoked via:

            * a **class**, the **class-wide** quality is unset.
            * an **instance**, the **instance-specific** quality is unset.

        If:

        * *value* < ``0``; JPEG encoding is disabled.
        * ``0`` <= *value* <= ``95``; JPEG encoding is enabled with the given quality.

        If **unset** for:

        * a **class**, it uses that of its parent *iterm2* style class (if any) or the
          default (disabled), if unset for all parents or the class has no parent
          *iterm2* style class.
        * an **instance**, it uses that of its class.

        By **default**, the quality is **unset** (i.e JPEG encoding is **disabled**) and
        images are encoded in the PNG format (when not reading directly from file) but
        in some cases, higher and/or faster compression may be desired.
        JPEG encoding is significantly faster than PNG encoding and produces smaller
        (in data size) output but **at the cost of image quality**.

        NOTE:
            * This property is :term:`descendant`.
            * This optimization applies to only **re-encoded** (i.e not read directly
              from file) **non-transparent** renders.

        TIP:
            The transparency status of some images can not be correctly determined
            in an efficient way at render time. To ensure JPEG encoding is always used
            for a re-encoded render, disable transparency or set a background color.

            Furthermore, to ensure that renders with the **WHOLE** :term:`render method`
            are always re-encoded, disable :py:attr:`read_from_file`.

            This optimization is useful in improving non-native animation performance.

        SEE ALSO:
            * the *alpha* parameter of :py:meth:`~term_image.image.BaseImage.draw`
              and the ``#``, ``bgcolor`` fields of the :ref:`format-spec`
            * :py:attr:`read_from_file`
        """,
    )

    native_anim_max_bytes = ClassProperty(
        lambda self: type(self)._native_anim_max_bytes,
        doc="""Maximum size (in bytes) of image data for native animation

        :type: int

        GET:
            Returns the set value.

        SET:
            A positive integer; the value is set.

            Can not be set via an instance.

        DELETE:
            The value is reset to the default.

            Can not be reset via an instance.

        :py:class:`~term_image.exceptions.TermImageWarning` is issued (and shown
        **only the first time**, except the warning filters are modified to do
        otherwise) if the image data size for a native animation is above this value.

        NOTE:
            This property is a global setting. Hence, setting/resetting it on this
            class or any subclass affects all classes and their instances.

        WARNING:
            This property should be altered with caution to avoid excessive memory
            usage, particularly on the terminal emulator's end.
        """,
    )

    read_from_file = ClassInstanceProperty(
        ITerm2ImageMeta.read_from_file.fget,
        ITerm2ImageMeta.read_from_file.fset,
        ITerm2ImageMeta.read_from_file.fdel,
        doc="""Read-from-file optimization

        :type: bool

        GET:
            Returns the effective read-from-file policy of the invoker
            (class or instance).

        SET:
            If invoked via:

            * a **class**, the **class-wide** policy is set.
            * an **instance**, the **instance-specific** policy is set.

        DELETE:
            If invoked via:

            * a **class**, the **class-wide** policy is unset.
            * an **instance**, the **instance-specific** policy is unset.

        If the value is:

        * ``True``, image data is read directly from file when possible and no image
          manipulation is required.
        * ``False``, images are always re-encoded (in the PNG format by default).

        If **unset** for:

        * a **class**, it uses that of its parent *iterm2* style class (if any) or the
          default (``True``), if unset for all parents or the class has no parent
          *iterm2* style class.
        * an **instance**, it uses that of its class.

        By **default**, the policy is **unset**, which is equivalent to ``True``
        i.e the optimization is **enabled**.

        NOTE:
            * This property is :term:`descendant`.
            * This is an optimization to reduce render times and is only applicable to
              the **WHOLE** render method, since the the **LINES** method inherently
              requires image manipulation.
            * This property does not affect animations. Native animations are always
              read from file when possible and frames of non-native animations have
              to be re-encoded.

        SEE ALSO:
            :py:attr:`jpeg_quality`
        """,
    )

    @classmethod
    def clear(cls, cursor: bool = False, now: bool = False) -> None:
        """Clears images.

        Args:
            cursor: If ``True``, all images intersecting with the current cursor
              position are cleared. Otherwise, all visible images are cleared.
            now: If ``True`` the images are cleared immediately, without affecting
              any standard I/O stream.
              Otherwise they're cleared when next :py:data:`sys.stdout` is flushed.

        NOTE:
            Required and works only on Konsole, as text doesn't overwrite images.
        """
        if not isinstance(cursor, bool):
            raise arg_type_error("cursor", cursor)
        if not isinstance(now, bool):
            raise arg_type_error("now", now)

        # There's no point checking for forced support since this is only required on
        # konsole which supports the protocol.
        # `is_supported()` is first called to ensure `_TERM` has been set.
        if cls.is_supported() and cls._TERM == "konsole":
            # Konsole utilizes the same image rendering implementation as it
            # uses for the kiity graphics protocol.
            (write_tty if now else _stdout_write)(
                (ctlseqs.KITTY_DELETE_CURSOR_b if now else ctlseqs.KITTY_DELETE_CURSOR)
                if cursor
                else (ctlseqs.KITTY_DELETE_ALL_b if now else ctlseqs.KITTY_DELETE_ALL)
            )

    @classmethod
    def is_supported(cls):
        if cls._supported is None:
            cls._supported = False

            name, version = get_terminal_name_version()
            if name in {"iterm2", "konsole", "wezterm"}:
                try:
                    if name != "konsole" or (
                        tuple(map(int, version.split("."))) >= (22, 4, 0)
                    ):
                        cls._supported = True
                        cls._TERM, cls._TERM_VERSION = name, version
                except ValueError:  # version string not "understood"
                    pass

        return cls._supported

    @classmethod
    def _check_style_format_spec(cls, spec: str, original: str) -> Dict[str, Any]:
        parent, (method, mix, compress) = cls._get_style_format_spec(spec, original)
        args = {}
        if parent:
            args.update(super()._check_style_format_spec(parent, original))
        if method:
            args["method"] = {"L": LINES, "W": WHOLE, "A": ANIM}[method]
        if mix:
            args["mix"] = bool(int(mix[-1]))
        if compress:
            args["compress"] = int(compress[-1])

        return cls._check_style_args(args)

    def _display_animated(
        self,
        img,
        alpha,
        fmt,
        *args,
        mix: bool = False,
        **kwargs,
    ):
        if not mix and self._TERM == "wezterm":
            lines = max(fmt[-1], self.rendered_height)
            r_width = self.rendered_width
            erase_and_move_cursor = ERASE_CHARS % r_width + CURSOR_FORWARD % r_width
            first_frame = self._format_render(
                f"{erase_and_move_cursor}\n" * (lines - 1) + erase_and_move_cursor,
                *fmt,
            )
            print(
                first_frame,
                "\r",
                CURSOR_UP % (lines - 1),
                sep="",
                end="",
                flush=True,
            )

        super()._display_animated(img, alpha, fmt, *args, mix=True, **kwargs)

    @staticmethod
    def _handle_interrupted_draw():
        """Performs necessary actions when image drawing is interrupted.

        If drawing is interrupted while transmitting an image, it causes terminal to
        wait for more data (while consuming any output following) until the output
        reaches the expected payload size or ST (String Terminator) is written.
        """

        # End last transmission (does no harm if there wasn't an unterminated
        # transmission)
        # Konsole sometimes requires ST to be written twice.
        print(ctlseqs.ST * 2, end="", flush=True)

    def _render_image(
        self,
        img: PIL.Image.Image,
        alpha: Union[None, float, str],
        *,
        frame: bool = False,
        method: Optional[str] = None,
        mix: bool = False,
        compress: int = 4,
    ) -> str:
        # NOTE: It's more efficient to write separate strings to the buffer separately
        # than concatenate and write together.

        # Using `width=<columns>`, `height=<lines>` and `preserveAspectRatio=0` ensures
        # that an image always occupies the correct amount of columns and lines even if
        # the cell size has changed when it's drawn.
        # Since we use `width` and `height` control data keys, there's no need
        # upscaling the image on this end to reduce payload.
        # Anyways, this also implies that the image(s) have to be resized by the
        # terminal emulator, thereby leaving various details of resizing in the hands
        # of the terminal emulator such as the resampling method, etc.
        # This particularly affects the LINES render method negatively, resulting in
        # slant/curved edges not lining up across lines (amongst other artifacts
        # observed on Konsole) supposedly because the terminal emulator resizes each
        # line separately.
        # Hence, this optimization is only used for the WHOLE render method.

        r_width, r_height = self.rendered_size
        render_method = (method or self._render_method).lower()

        # Workarounds
        is_on_konsole = self._TERM == "konsole"
        is_on_wezterm = self._TERM == "wezterm"
        cursor_right = CURSOR_FORWARD % r_width
        cursor_up = CURSOR_UP % (r_height - 1) if r_height > 1 else ""
        erase = ERASE_CHARS % r_width if not mix and is_on_wezterm else ""

        file_is_readable = True
        if self._source_type is ImageSource.PIL_IMAGE:
            try:
                file_is_readable = os.access(img.filename, os.R_OK)
            except (AttributeError, OSError):
                file_is_readable = False

        if render_method == ANIM and self._is_animated and not frame:
            if self._source_type is ImageSource.PIL_IMAGE:
                if file_is_readable:
                    compressed_image = open(img.filename, "rb")
                else:
                    compressed_image = io.BytesIO()
                    try:
                        img.save(compressed_image, img.format, save_all=True)
                    except ValueError as e:
                        self._close_image(img)
                        raise RenderError(
                            "iTerm2 native animation not supported: This image was "
                            "sourced from a PIL image with an unknown format"
                        ) from e
            else:
                compressed_image = open(self._source, "rb")

            self._close_image(img)

            with compressed_image:
                compressed_image.seek(0, 2)
                if compressed_image.tell() > self.native_anim_max_bytes:
                    warnings.warn(
                        "Image data size above the maximum for native animation",
                        TermImageWarning,
                    )

                control_data = "".join(
                    (
                        f"size={compressed_image.tell()};width={r_width}"
                        f";height={r_height};preserveAspectRatio=0;inline=1"
                        f"{';doNotMoveCursor=1' * is_on_konsole}:"
                    )
                )
                compressed_image.seek(0)
                return "".join(
                    (
                        (
                            ""
                            if is_on_konsole
                            else f"{erase}{cursor_right}\n" * (r_height - 1)
                        ),
                        erase,
                        "" if is_on_konsole else cursor_up,
                        ITERM2_START,
                        control_data,
                        standard_b64encode(compressed_image.read()).decode(),
                        ST,
                        f"{cursor_right}\n" * (r_height - 1) if is_on_konsole else "",
                        cursor_right * is_on_konsole,
                    )
                )

        width, height = (
            self._get_minimal_render_size()
            if render_method == WHOLE
            else self._get_render_size()
        )

        if (  # Read directly from file when possible and reasonable
            self.read_from_file
            and not self._is_animated
            and file_is_readable
            and render_method == WHOLE
            and mul(*self._original_size) <= mul(*self._get_render_size())
            and (
                # None of the *alpha* options can affect these
                img.mode in {"1", "L", "RGB", "HSV", "CMYK"}
                # Alpha threshold is unused with graphics-based styles.
                # The transparency of some "P" mode images is missing on some terminals
                # Making the output inconsistent with other render styles.
                or (isinstance(alpha, float) and img.mode not in {"P", "PA"})
            )
        ):
            compressed_image = open(
                img.filename
                if self._source_type is ImageSource.PIL_IMAGE
                else self._source,
                "rb",
            )
            frame_img = None
        else:
            frame_img = img if frame else None
            img = self._get_render_data(
                img, alpha, size=(width, height), pixel_data=False, frame=frame
            )[0]  # fmt: skip
            if self.jpeg_quality >= 0 and img.mode == "RGB":
                format = "jpeg"
                jpeg_quality = self.jpeg_quality
            else:
                format = "png"
                jpeg_quality = None

            if render_method == LINES:
                raw_image = io.BytesIO(img.tobytes())
                compressed_image = io.BytesIO()
            else:
                compressed_image = io.BytesIO()
                img.save(
                    compressed_image,
                    format,
                    compress_level=compress,  # PNG
                    quality=jpeg_quality,
                )

        # clean up (ImageIterator uses one PIL image throughout)
        if frame_img is not img:
            self._close_image(img)

        if render_method == LINES:
            # NOTE: It's more efficient to write separate strings to the buffer
            # separately than concatenate and write together.

            cell_height = height // r_height
            bytes_per_line = width * cell_height * (len(img.mode))
            control_data = (
                f";width={r_width};height=1;preserveAspectRatio=0;inline=1"
                f"{';doNotMoveCursor=1' * is_on_konsole}:"
            )

            with io.StringIO() as buffer, raw_image, compressed_image:
                for line in range(1, r_height + 1):
                    compressed_image.seek(0)
                    with PIL.Image.frombytes(
                        img.mode, (width, cell_height), raw_image.read(bytes_per_line)
                    ) as img:
                        img.save(
                            compressed_image,
                            format,
                            compress_level=compress,  # PNG
                            quality=jpeg_quality,
                        )
                    compressed_image.truncate()

                    buffer.write(erase)
                    buffer.write(ITERM2_START)
                    buffer.write(f"size={compressed_image.tell()}")
                    buffer.write(control_data)
                    buffer.write(
                        standard_b64encode(compressed_image.getvalue()).decode()
                    )
                    buffer.write(ST)
                    is_on_konsole and buffer.write(cursor_right)
                    line < r_height and buffer.write("\n")

                return buffer.getvalue()

        # WHOLE
        with compressed_image:
            compressed_image.seek(0, 2)
            control_data = "".join(
                (
                    f"size={compressed_image.tell()};width={r_width}"
                    f";height={r_height};preserveAspectRatio=0;inline=1"
                    f"{';doNotMoveCursor=1' * is_on_konsole}:"
                )
            )
            compressed_image.seek(0)
            return "".join(
                (
                    (
                        ""
                        if is_on_konsole
                        else f"{erase}{cursor_right}\n" * (r_height - 1)
                    ),
                    erase,
                    "" if is_on_konsole else cursor_up,
                    ITERM2_START,
                    control_data,
                    standard_b64encode(compressed_image.read()).decode(),
                    ST,
                    f"{cursor_right}\n" * (r_height - 1) if is_on_konsole else "",
                    cursor_right * is_on_konsole,
                )
            )


_stdout_write = sys.stdout.write
